//package gliparameterizer;

import java.util.Vector;

/**
 * Class for doing GLI calculations.
 * <p>Copyright: Copyright (c) 2004 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */

class GLICalcs {
  public GLICalcs() {
  }

  /**
   * Finds the trees that intersect a given ray.
   * @param oRay GLIRay Ray to find intersecting trees for.
   * @param oAllTrees Tree collection.
   * @return Vector Vector of trees.  The size will be 0 if there are
   * no intersectors.
   */
  public static Vector FindIntersectors(GLIRay oRay, AllTrees oAllTrees) {
    Vector oIntersectors = new Vector(0);

    //Get all trees that could possibly intersect
    Vector oPossibleIntersectors = oAllTrees.FindTrees(oRay);
    Tree oTree;
    double fDistance;
    int i;

    //Test each one to see if they do intersect
    for (i = 0; i < oPossibleIntersectors.size(); i++) {
      oTree = (Tree) oPossibleIntersectors.elementAt(i);

      //Does the canopy cover the GLI point?
      fDistance = GetDistance(oRay.m_fX, oRay.m_fY, oTree.m_fX, oTree.m_fY);
      if (fDistance - oTree.m_fCanopyRadius < 0) {
        if (GLICanopyIntersect(oRay, oTree)) {
          oIntersectors.add(oTree);
        }
      } else {
        if (GLIIntersect(oRay, oTree)) {
          oIntersectors.add(oTree);
        }
      }
    }

    return oIntersectors;
  }

  /**
   * Determines whether or not a tree intersects a given ray.
   * @param oRay Ray to test for intersection
   * @param oNeighbor Neighbor tree for assessing ray interception
   * @return boolean True if the tree intersects the ray, false if not.
   */
  public static boolean GLIIntersect(GLIRay oRay, Tree oNeighbor) {

    double fSquareOfNearDist, //square of the distance to the nearest point
        //that is a solution to the system of equations
        fSquareOfFarDist, //square of the distance to the farthest point
        //that is a solution to the system of equations
        fFarX, fNearX, fFarY, fNearY, //two solutions of system of equations -
        //for the far and near sides of neighbor crown
        fA, fB, fC, //variables in quadratic formula
        fXDist, fYDist, //distance between two trees in X and Y planes
        fHeightDiff, //difference in crown top heights
        fHeightDiffSquared, //squared difference in heights
        fCrownBaseDiff, //difference in height between neighbor crown
        //base and target tree crown top
        fAzimuth, //azimuth to neighbor
        fCrownAngle, //angle subtended by half of crown
        fCrownBaseDiffSquared, //squared difference between crown base and ht.
        fAltStart, //lower altitude angle blocked by neighbor
        fAltEnd, //upper altitude angle blocked by neighbor
        fGLISlope; //slope of azimuth of ray

    //**************************************
    // Azimuth
    //**************************************
    //Distance between the trees in the X and Y planes
    fXDist = oNeighbor.m_fX - oRay.m_fX;
    fYDist = oNeighbor.m_fY - oRay.m_fY;

    //Solve the system of equations of the circle of the neighbor's canopy
    //and the azimuth from the GLI point.  Do all calculations as though the
    //GLI point is at the origin.

    //Slope of the line made by the ray
    fGLISlope = java.lang.Math.tan(oRay.m_fAzimuth);
    if (fGLISlope == 0) fGLISlope = 10000;
    else fGLISlope = 1/fGLISlope;

    //Plug in for a, b, and c in the quadratic formula
    fA = 1.0 + fGLISlope * fGLISlope;
    fB = -2.0 * (fXDist + fGLISlope * fYDist);
    fC = fYDist * fYDist + fXDist * fXDist - oNeighbor.m_fCanopyRadius * oNeighbor.m_fCanopyRadius;
    fC = fB * fB - 4.0 * fA * fC;

    //If fC is negative, it means that the system of equations does not have
    //a real solution, and thus the circle and line don't intersect.
    if (fC < 0.0) {
      return false;
    }

    //The slope defines a line that extends in both directions; we're only
    //interested in a ray extending in one direction.  Eliminate those trees
    //which are 180 degrees from our azimuth.
    fAzimuth = GetAzimuthAngle(oRay.m_fX, oRay.m_fY, oNeighbor.m_fX, oNeighbor.m_fY);
    fCrownAngle = java.lang.Math.asin(oNeighbor.m_fCanopyRadius /
                                      GetDistance(oRay.m_fX, oRay.m_fY, oNeighbor.m_fX, oNeighbor.m_fY));
    if (oRay.m_fAzimuth > (fAzimuth + fCrownAngle) ||
        oRay.m_fAzimuth < (fAzimuth - fCrownAngle)) {
      return false;
    }

    //**************************************
    // Altitude
    //**************************************
    //Finish solving the quadratic formula so we can calculate altitude angles.
    //We'll have an X,Y point for the slope's intersection with the near and
    //far sides of the neighbor's crown
    fC = java.lang.Math.sqrt(fC);
    fNearX = ( -fB - fC) / (2.0 * fA);
    fFarX = ( -fB + fC) / (2.0 * fA);
    fNearY = fGLISlope * fNearX;
    fFarY = fGLISlope * fFarX;

    //Calculate the squares of the distances to the near and far points on the
    //neighbor's crown - we don't need to bother taking the square roots
    fSquareOfNearDist = fNearX * fNearX + fNearY * fNearY;
    fSquareOfFarDist = fFarX * fFarX + fFarY * fFarY;

    //Switch them in case we were wrong about which point was near
    if (fSquareOfNearDist > fSquareOfFarDist) {
      fA = fSquareOfNearDist; //dangerously reusing fA!
      fSquareOfNearDist = fSquareOfFarDist;
      fSquareOfFarDist = fA;
    }

    //Calculate the starting altitude
    fCrownBaseDiff = oNeighbor.m_fCanopyBottomHeight - oRay.m_fHeight;
    fCrownBaseDiffSquared = fCrownBaseDiff * fCrownBaseDiff;
    fAltStart = java.lang.Math.asin(fCrownBaseDiff /
                 java.lang.Math.sqrt( (fSquareOfFarDist + fCrownBaseDiffSquared)));

    if (fAltStart > oRay.m_fAltitude) {
      return false;
    }

    //Calculate the ending altitude
    fHeightDiff = oNeighbor.m_fHeight - oRay.m_fHeight;
    fHeightDiffSquared = fHeightDiff * fHeightDiff;
    fAltEnd = java.lang.Math.asin( (oNeighbor.m_fHeight - oRay.m_fHeight) /
               java.lang.Math.sqrt( (double) (fSquareOfNearDist +
                                              fHeightDiffSquared)));
    if (fAltEnd < oRay.m_fAltitude) {
      return false;
    }

    return true;
  }

  /**
   * Tells if a tree canopy intersects a given ray, if the origin of the ray
   * is under the tree canopy.
   * @param oRay GLIRay Ray to test for intersection.
   * @param oNeighbor Tree Tree to test for intersection.
   * @return boolean True if the ray intersects; false if it doesn't.
   */
  public static boolean GLICanopyIntersect(GLIRay oRay, Tree oNeighbor) {
   double fXDist, fYDist, //distance between two trees in X and Y directions
         fGLISlope, //slope of the line made by the azimuth angle
         fAzimuth, //azimuth to a point on the neighbor's canopy
         fDistToFarEdge,    //square of distance to farthest edge of neighbor
         fX1, fX2, fY1, fY2, //two solutions of system of equations
         fAltStart, //lowest altitude of neighbor crown
         fHeightDiff,    //difference in target height and crown base
         fA, fB, fC,     //variables in quadratic formula
         fCCoeff;        //one quadratic coefficient that can be done ahead

   fXDist = oNeighbor.m_fX - oRay.m_fX;
   fYDist = oNeighbor.m_fY - oRay.m_fY;

   if (oNeighbor.m_fCanopyBottomHeight <= oRay.m_fHeight) {//neighbor blocks whole sky
     return true;

   } else { //neighbor blocks part of sky

       //we have to use linear algebra to solve the system of equations of
       //the circle of the neighbor's canopy and the line from the target to
       //the center of the azimuth chunk in question.  The slope of the line
       //is in mp_fAziSlope.  Do all calculations as though the target tree
       //is at the origin.

       //Slope of the line made by the ray
       fGLISlope = java.lang.Math.tan(oRay.m_fAzimuth);
       if (fGLISlope == 0) fGLISlope = 10000;
       else fGLISlope = 1/fGLISlope;

       //Solve the quadratic formula
       fA = 1.0 + fGLISlope * fGLISlope;
       fB = -2.0 * (fXDist + fGLISlope * fYDist);
       fCCoeff = (fXDist*fXDist) + (fYDist*fYDist) - (oNeighbor.m_fCanopyRadius * oNeighbor.m_fCanopyRadius);
       fC = fB * fB - 4.0 * fA * fCCoeff;
       if (fC < 0) fC = 0; //test for sqrt bug
       fC = java.lang.Math.sqrt((double) fC);

       //Get the solutions to find the intersection points between the circle
       //and line - these points are (fX1, fY1) and (fX2, fY2)
       fX1 = (-fB + fC) / (2.0 * fA);
       fX2 = (-fB - fC) / (2.0 * fA);
       fY1 = fGLISlope * fX1;
       fY2 = fGLISlope * fX2;

       //We've solved for the intersection of a circle and a line, but we
       //really want the solution to the intersection of a circle and a ray.
       //Figure out which solution represents the ray, and which is the
       //negative ray.
       fAzimuth = GetAzimuthAngle(0, 0, fX1, fY1);
       if (java.lang.Math.abs(fAzimuth - oRay.m_fAzimuth) > 0.01) {
         //It's not fX1 and fY1 - so assign the other point to them so that
         //fX1 and fY1 hold the point we want
         fX1 = fX2;
         fY1 = fY2;
       }

       //For this point, calculate the lowest altitude of the canopy
       fDistToFarEdge = java.lang.Math.sqrt((fX1*fX1) + (fY1*fY1));
       fHeightDiff = oNeighbor.m_fCanopyBottomHeight - oRay.m_fHeight;
       fAltStart = java.lang.Math.atan(fHeightDiff / fDistToFarEdge);
       if (fAltStart <= oRay.m_fAltitude) return true;
   } //end of else
   return false;
}

  /**
   * Gets the distance between 2 points.  No torus wrapping.
   * @param fFromX X coordinate of the "from" point.
   * @param fFromY Y coordinate of the "from" point.
   * @param fToX X coordinate of the "to" point.
   * @param fToY Y coordinate of the "to" point.
   * @return Distance between the two points in meters.
   */
  public static double GetDistance(double fFromX, double fFromY, double fToX, double fToY) {

    double fXDistance, fYDistance;

    fXDistance = fToX - fFromX;
    fYDistance = fToY - fFromY;
    return java.lang.Math.sqrt( (fXDistance * fXDistance) +
                               (fYDistance * fYDistance));
  }

  /**
   * Gets the azimuth angle between two points, in radians.
   * @param fFromX X coordinate of the "from" point.
   * @param fFromY Y coordinate of the "from" point.
   * @param fToX X coordinate of the "to" point.
   * @param fToY Y coordinate of the "to" point.
   * @return double Azimuth angle in radians.
   */
  public static double GetAzimuthAngle(double fFromX, double fFromY, double fToX, double fToY) {

    double fDistX = fToX - fFromX,
        fDistY = fToY - fFromY,
        fAzimuth = 0; //azimuth angle to return

    //Calculate the azimuth - correct for quadrant of "to" relative to "from"
    //(counting clockwise from upper right)
    if ( (fDistY > 0) && (fDistX >= 0)) { //first quadrant
      fAzimuth = java.lang.Math.atan( (fDistX) / (fDistY));
    }
    if ( (fDistX > 0) && (fDistY <= 0)) { //second quadrant
      fAzimuth = (java.lang.Math.PI / 2.0) + java.lang.Math.atan( ( -1.0 * fDistY) / (fDistX));
    }
    if ( (fDistX <= 0) && (fDistY < 0)) { //third quadrant
      fAzimuth = java.lang.Math.PI + java.lang.Math.atan( ( -1.0 * fDistX) / ( -1.0 * fDistY));
    }
    if ( (fDistX < 0) && (fDistY >= 0)) { //fourth quadrant
      fAzimuth = (1.5 * java.lang.Math.PI) + java.lang.Math.atan( (fDistY) / ( -1.0 * fDistX));
    }
    return fAzimuth;
  }

}
